# libraries
from datetime import datetime, timedelta

import numpy as np
import pandas as pd


def format_tank(dict_input, freq: str = "30min"):

    if dict_input is None:
        return None

    validated_input = []

    for i in range(len(dict_input)):
        if "read_time" in dict_input[i] and "volume" in dict_input[i]:
            validated_input.append(dict_input[i])

    if validated_input == []:
        return None

    validated_input = sorted(validated_input, key=lambda x: x["read_time"])
    validated_input = [d for d in validated_input if d["volume"]]
    df = pd.DataFrame(validated_input)
    df.read_time = pd.to_datetime(df.read_time)
    df["y"] = df.volume.diff() * -1
    df = df[["read_time", "y"]].rename(columns={"read_time": "ds"})
    df = df[df.y < 800].copy()
    df.y = df.y.clip(lower=0)
    df = df.set_index("ds").groupby(pd.Grouper(freq=freq)).sum()

    if len(df) == 2:
        return None

    df.reset_index(level=0, inplace=True)
    output = df.iloc[1:-1]

    return output


def format_sales(dict_input):

    if dict_input == [] or dict_input is None:
        return None

    dict_input = sorted(dict_input, key=lambda x: x["date"])
    dat = pd.DataFrame(dict_input)
    df = dat[["date", "sales"]]

    df2 = df.copy()
    df2["sales"] = df["sales"].clip(
        lower=df.sales.quantile(0.015), upper=df.sales.quantile(0.995)
    )

    df3 = df2.rename(columns={"date": "ds", "sales": "y"})

    df4 = df3.copy()
    df4["ds"] = df3.loc[:, "ds"].astype("datetime64[ns]")

    return df4


def gen_past(df, freq="30min", periods=432):

    if len(df) == 0 or df is None:
        now = datetime.now()
        end_0 = now - (now - datetime.min) % timedelta(minutes=30)

    else:
        end_0 = df.iloc[-1, 0]

    past = pd.date_range(end=end_0, freq=freq, periods=periods)
    past = pd.DataFrame(past)
    past.rename(columns={0: "ds"}, inplace=True)
    past["ds"] = pd.to_datetime(past["ds"], format="%Y-%m-%d")

    return past


def validate_tank(tank_data, sales_data):

    if len(tank_data) == 0 or tank_data is None:
        return None

    # add day of year for tank data / sales data
    sales_data["doy"] = sales_data["ds"].dt.dayofyear
    sales_data = sales_data.iloc[-100:]

    # create validation dataframe
    val_matrix = np.zeros((432, 2))
    val_df = pd.DataFrame(val_matrix, columns=["ds", "doy"])

    # dynamically generate timestamps
    val_df["ds"] = gen_past(tank_data, freq="30min", periods=432)
    val_df["doy"] = val_df.ds.dt.dayofyear
    val_df["ts"] = val_df.ds.dt.time

    val_df = pd.merge(
        left=val_df, right=generic_daily_decomp, how="left", left_on="ts", right_on="ts"
    )
    val_df = pd.merge(
        left=val_df, right=tank_data, how="left", left_on="ds", right_on="ds"
    )
    val_df = pd.merge(
        left=val_df, right=sales_data, how="left", left_on="doy", right_on="doy"
    )

    val_df.rename(columns={"ds_x": "ds", "y_x": "actual", "y_y": "daily"}, inplace=True)
    val_df = val_df.loc[:, ["ds", "ts", "actual", "daily", "daily_multi"]]

    # calculate generalized daily tank sticking curve from sales data
    val_df["base_est"] = val_df.loc[:, "daily"] * (1 / 48)
    est = val_df["base_est"] * val_df["daily_multi"]
    val_df["estimated"] = est

    val_df = val_df.loc[:, ["ds", "actual", "estimated"]]

    # data abnormalities adjustment
    diff = val_df.estimated - val_df.actual
    threshold = diff.abs().quantile(0.95)

    mask_adj = diff.abs() > threshold
    mask_tank = [not i for i in mask_adj]

    # val_df.loc[mask_adj, 'output'] = val_df.loc[mask_adj, 'adj_sales']
    # val_df.loc[mask_tank, 'output'] = val_df.loc[mask_tank, 'tank_vals']

    # missing data imputation
    mask_nan = val_df.actual.isna()
    mask_not_nan = [not i for i in mask_nan]

    val_df.loc[mask_nan, "y"] = val_df.loc[mask_nan, "estimated"]
    val_df.loc[mask_not_nan, "y"] = val_df.loc[mask_not_nan, "actual"]

    # for testing purposes, uncomment this
    # return val_df

    # for use in forecasting, uncomment this
    val_df.rename(columns={"ds_x": "ds"}, inplace=True)

    return val_df[["ds", "y"]]


if __name__ == "__main__":
    raw_readings = [
        {
            "id": "5f8de916228d65344402f0a7",
            "store_number": "110",
            "tank_id": "1",
            # "run_time": "2020-10-19T19:29:17",
            # "read_time": "2020-10-19T19:29:23",
            "volume": 13547.508587862303,
            "temperature": 48.1715720177702,
        },
        {
            "id": "5f8de914228d65344402f083",
            "store_number": "110",
            "tank_id": "1",
            "run_time": "2 020-10-19T19:29:16",
            # "read_time": "2020-10-19T19:29:21",
            # "volume": 13549.028743568035,
            "temperature": 48.1710855651926,
        },
        {
            "id": "5f8de49c1c6cbdc17a3546f9",
            "store_number": "110",
            "tank_id": "1",
            # "run_time": "2020-10-19T19:10:13",
            # "read_time": "2020-10-19T19:10:18",
            "volume": 13621.377813533778,
            "temperature": 48.0932317562486,
        },
        {
            "id": "5f8de49b1c6cbdc17a3546e5",
            "store_number": "110",
            "tank_id": "1",
            "run_time": "2020-10-19T19:10:13",
            "read_time": "2020-10-19T19:10:16",
            # "volume": 13621.377813533778,
            "temperature": 48.0932317562486,
        },
        {
            "id": "5f8de0011dff1f3a8c927fc1",
            "store_number": "110",
            "tank_id": "1",
            "run_time": "2020-10-19T18:50:35",
            "read_time": "2020-10-19T18:50:37",
            # "volume": 13792.582207190737,
            "temperature": 48.025463112635,
        },
    ]
    format_tank(raw_readings)



from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import requests
from fbprophet import Prophet
from sparkles.fbprophet_hacks import suppress_stdout_stderr

# declare variables for testing
store = "KT220"
tank = "7"
fc_url = "https://fc.bbdev.host3.capspire.com"


def decompose_sales(sales_data):

    # not enough data
    if sales_data is None or len(sales_data) < 7:
        return pd.Series([1] * 53), pd.Series([1] * 7)

    # fit model for weekly seasonality
    m = Prophet(
        changepoint_prior_scale=0.05,
        daily_seasonality=False,
        weekly_seasonality=True,
        yearly_seasonality=True,
        seasonality_mode="multiplicative",
    )
    with suppress_stdout_stderr():
        m.fit(sales_data)

    # calculate weekly seasonality
    future = m.make_future_dataframe(periods=1, freq="1D")
    forecast = m.predict(future)
    forecast_reduced = forecast.loc[:, ["ds", "yearly", "weekly"]]
    forecast_reduced["dow"] = forecast_reduced.ds.dt.weekday
    forecast_reduced["doy"] = forecast_reduced.ds.dt.dayofyear
    forecast_reduced["woy"] = forecast_reduced.ds.dt.week

    weekly_seasonality = forecast_reduced.groupby("dow")["weekly"].mean() + 1

    if len(weekly_seasonality) < 7:
        return pd.Series([1] * 53), pd.Series([1] * 7)

    if len(sales_data) < 350:
        return pd.Series([1] * 53), weekly_seasonality

    # if nearly a full year, or more of data exists, calculate yearly seasonality
    forecast_reduced = forecast_reduced.iloc[-366:]
    forecast_reduced.sort_values(by=["woy", "dow"], inplace=True)
    year_trend = forecast_reduced.groupby("woy").mean()
    year_trend = year_trend["yearly"] + 1

    yearly_seasonality = ensure_53_weeks(year_trend)

    if len(weekly_seasonality) == 7 and len(yearly_seasonality) == 53:
        return yearly_seasonality, weekly_seasonality

    return pd.Series([1] * 53), pd.Series([1] * 7)


def ensure_53_weeks(initial: pd.DataFrame) -> pd.DataFrame:
    """
    takes a shorter dataframe and ensures at least 53 rows by wrapping the first rows onto the end
    """
    extra_weeks_needed = 53 - len(initial)
    if extra_weeks_needed > 0:
        ret = pd.concat([initial, initial[:extra_weeks_needed]], ignore_index=True)
    else:
        ret = initial.copy()
    return ret


def decompose_tank(tank_history):

    if tank_history is None or len(tank_history) < 48:
        return generic_hh_seasonality

    m = Prophet(
        changepoint_prior_scale=0.05,
        daily_seasonality=True,
        weekly_seasonality=True,
        yearly_seasonality=False,
        seasonality_mode="multiplicative",
    )
    with suppress_stdout_stderr():
        m.fit(tank_history)

    future = m.make_future_dataframe(periods=1, freq="30min")
    forecast = m.predict(future)

    forecast_reduced = forecast.loc[:, ["ds", "daily"]]
    forecast_reduced["time"] = forecast_reduced["ds"].dt.time
    forecast_reduced.drop(columns="ds", inplace=True)
    day_trend = forecast_reduced.groupby("time").mean()
    day_trend.sort_values(by=["time"], inplace=True)

    if len(day_trend) != 48:
        return generic_hh_seasonality

    hh_seasonality = day_trend["daily"] + 1

    return hh_seasonality


def decompose_holiday(input):
    if type(input) != pd.core.frame.DataFrame:
        return print("error - input is not a DataFrame")
    else:
        event_1 = pd.DataFrame(
            {
                "holiday": "event_1_name",
                "ds": pd.to_datetime(["2020-03-01"]),
                "lower_window": 0,
                "upper_window": 0,
            }
        )
        event_2 = pd.DataFrame(
            {
                "holiday": "event_2_name",
                "ds": pd.to_datetime(["2020-03-02"]),
                "lower_window": 0,
                "upper_window": 0,
            }
        )
        holidays = pd.concat([event_1, event_2])
        m = Prophet(
            holidays=holidays,
            changepoint_prior_scale=0.05,
            daily_seasonality=False,
            weekly_seasonality=True,
            yearly_seasonality=True,
            seasonality_mode="multiplicative",
        )
        m.add_country_holidays(country_name="US")
        with suppress_stdout_stderr():
            m.fit(input)
        future = m.make_future_dataframe(periods=1, freq="1D")
        forecast = m.predict(future)
        forecast_reduced = forecast.loc[:, ["ds", "holidays"]]

    return forecast_reduced


generic_weekly_seasonality = [0.95, 1.0, 1.05, 1.05, 1.2, 0.95, 0.8]
generic_yearly_seasonality = [
    0.95,
    0.95,
    *np.repeat(0.9, 10),
    0.925,
    0.95,
    0.975,
    *np.repeat(1, 7),
    1.025,
    1.05,
    *np.repeat(1.075, 11),
    1.05,
    1.025,
    *np.repeat(1, 7),
    *np.repeat(0.95, 9),
]
generic_hh_seasonality = [
    1.0,
    0.85,
    0.7,
    0.55,
    0.4,
    0.25,
    0.1,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.0,
    0.1,
    0.2,
    0.3,
    0.4,
    0.5,
    0.6,
    0.7,
    0.8,
    0.9,
    1.0,
    1.1,
    1.2,
    1.3,
    1.4,
    1.5,
    1.6,
    1.7,
    1.8,
    1.9,
    2.0,
    2.0,
    2.0,
    2.0,
    1.9,
    1.8,
    1.7,
    1.6,
    1.5,
    1.4,
    1.3,
    1.2,
    1.1,
]


# request sales data
# def get_sales_history(
#     store_number: str, tank_id: str,
# ):
#     parameters = {
#         "store_number": store_number,
#         "tank_id": tank_id,
#         "limit": 365,
#         "system_psk": "lol",
#     }
#
#     r = requests.post(url=f"{fc_url}/sales_data/list", params=parameters)
#
#     return r.json()


from tank_forecaster import validation

#
# def main():
#     x = get_sales_history(store_number=store, tank_id=tank)
#     y = validation.format_sales(x)
#     z = decompose_sales(y)
#     print(z[0], z[1])
#     print(len(z[0]), len(z[1]))

tank_url = "https://bbdev.host3.capspire.com/api/tms"


def now_utc(exact: bool = False) -> datetime:
    import pytz

    date_time_now = datetime.now(pytz.utc).replace(tzinfo=None)
    if exact:
        return date_time_now
    return date_time_now.replace(second=0, microsecond=0)


def get_readings(
    store_number: str,
    tank_id: str,
    start_date: datetime = now_utc() - timedelta(days=16),
):
    parameters = {
        "store_number": store_number,
        "tank_id": tank_id,
        "start_date": start_date,
        "system_psk": "lol",
    }
    r = requests.post(url=f"{tank_url}/tank_inventory/readings", params=parameters)
    return r.json()


from tank_forecaster import validation


def main():
    x = get_readings(store_number=store, tank_id=tank)
    y = validation.format_tank(x)
    z = decompose_tank(y)
    print(z)


if __name__ == "__main__":
    main()
